import tensorflow as tf


def clean_acc(y_true, y_pred):
    labels = tf.case(
        [(tf.equal(tf.shape(y_true)[1], 1), 
            lambda: tf.cast(y_true, 'int32')[:,0])],
        lambda: tf.cast(tf.argmax(y_true, axis=1), 'int32'))

    return tf.reduce_mean(tf.cast(
        tf.equal(labels, tf.cast(tf.argmax(y_pred[:, :-1], axis=1), 'int32')),
        'float32'))


def vra(y_true, y_pred):
    labels = tf.case(
        [(tf.equal(tf.shape(y_true)[1], 1), 
            lambda: tf.cast(y_true, 'int32')[:,0])],
        lambda: tf.cast(tf.argmax(y_true, axis=1), 'int32'))

    return tf.reduce_mean(tf.cast(
        tf.equal(labels, tf.cast(tf.argmax(y_pred, axis=1), 'int32')),
        'float32'))



import tensorflow as tf
import tensorflow.keras.backend as K

from scipy.stats import truncnorm


def l2_normalize(x):
    return x / (tf.sqrt(tf.reduce_sum(x**2.)) + K.epsilon())


def np_truncated_normal(shape):
    return truncnorm.rvs(-2., 2., size=shape).astype('float32')


def global_lipschitz_bound_spectral_power(
        model, 
        num_iterations, 
        norm='l2', 
        exclude_last_layer=True, 
        maintain_state=True):

    k = 1.

    while_cond = lambda i, _: i < num_iterations

    layers = model.layers[:-1] if exclude_last_layer else model.layers

    # Used if we're maintaining the iterates' states.
    iterates = []
    updates = []

    for layer in layers:
        if not hasattr(layer, 'kernel'):
            continue

        W = layer.kernel

        if len(W.shape) == 4:
            # This is a convolutional layer.
            if norm == 'l2':
                if maintain_state:
                    x_0 = tf.Variable(
                        np_truncated_normal((1,*layer.input_shape[1:])),
                        trainable=False)

                    iterates.append(x_0)

                else:
                    x_0 = tf.random.truncated_normal(
                        shape=(1,*layer.input_shape[1:]))

                def body(i, x):
                    x = l2_normalize(x)

                    x_p = tf.nn.conv2d(
                        x,
                        W,
                        strides=layer.strides,
                        padding=layer.padding.upper())
                    x = tf.nn.conv2d_transpose(
                        x_p,
                        W,
                        x.shape,
                        strides=layer.strides,
                        padding=layer.padding.upper())

                    return i + 1, x

                _, x = tf.while_loop(while_cond, body, [tf.constant(0), x_0])

                if maintain_state:
                    updates.append((x_0, x))

                Wx = tf.nn.conv2d(
                    x, W, strides=layer.strides, padding=layer.padding.upper())

                k *= tf.sqrt(tf.reduce_sum(Wx**2.) / tf.reduce_sum(x**2.))

            else:
                raise ValueError(f'{norm} norm not supported')

        else:
            if norm == 'l2':
                if maintain_state:
                    x_0 = tf.Variable(
                        np_truncated_normal((W.shape[1], 1)),
                        trainable=False)

                    iterates.append(x_0)

                else:
                    x_0 = tf.random.truncated_normal(shape=(W.shape[1], 1))

                def body(i, x):
                    x = l2_normalize(x)
                    x_p = W @ x
                    x = tf.transpose(W) @ x_p

                    return i + 1, x

                _, x = tf.while_loop(while_cond, body, [tf.constant(0), x_0])

                if maintain_state:
                    updates.append((x_0, x))

                k *= tf.sqrt(tf.reduce_sum((W @ x)**2.) / tf.reduce_sum(x**2.))

            else:
                raise ValueError(f'{norm} norm not supported')

    if maintain_state:
        return k, iterates, updates

    return k



import tensorflow.keras.backend as K

from tensorflow.keras.models import Model

from gloro.layers import LipschitzMargin
from gloro.training.callbacks import UpdatePowerIterates


class GloroNet(Model):
    def __init__(
            self, 
            inputs=None, 
            outputs=None, 
            epsilon=None, 
            num_iterations=5,
            maintain_state=True,
            model=None, 
            hardcoded_kW=None, 
            **kwargs):

        if epsilon is None:
            raise ValueError('`epsilon` is required')
        
        if model is None and (inputs is None or outputs is None):
            raise ValueError(
                'must specify either `inputs` and `outputs` or `model`')
            
        if model is not None and inputs is not None and outputs is not None:
            raise ValueError(
                'cannot specify both `inputs` and `outputs` and `model`')
        
        if inputs is None or outputs is None:
            inputs = model.inputs
            outputs = model.outputs
            
        else:
            model = Model(inputs, outputs)
            
        if not isinstance(outputs, (list, tuple)):
            outputs = [outputs]
            
        if len(outputs) > 1:
            raise ValueError(
                f'only models with a single output are supported, but got '
                f'{len(outputs)} outputs')
        
        margin_layer = LipschitzMargin(
            epsilon, 
            model, 
            num_iterations=num_iterations, 
            maintain_state=maintain_state,
            hardcoded_kW=hardcoded_kW)

        margin = margin_layer(*outputs)
        
        super().__init__(inputs, margin, **kwargs)
        
        self._inputs = inputs
        self._outputs = outputs
        
        self._margin_layer = margin_layer
        self._f = GloroNet.__ModelContainer(model)

        self._lipschitz_constant_fn = K.function(
            self._inputs,
            [self._margin_layer._lipschitz_constant_tensor])

    @property
    def epsilon(self):
        return self._margin_layer.epsilon

    @epsilon.setter
    def epsilon(self, new_epsilon):
        self._margin_layer.epsilon = new_epsilon

    @property
    def num_iterations(self):
        return self._margin_layer.num_iterations

    @num_iterations.setter
    def num_iterations(self, new_num_iterations):
        self._margin_layer.num_iterations = new_num_iterations

    @property
    def maintain_state(self):
        return self._margin_layer.maintain_state

    @property
    def f(self):
        return self._f.model

    def lipschitz_constant(self, X):
        if not isinstance(X, (list, tuple)):
            X = [X]

        return self._lipschitz_constant_fn(X)[0]

    def predict_clean(self, *args, **kwargs):
        return self._f.model.predict(*args, **kwargs)

    def freeze_lipschitz_constant(self, converge=True):
        if converge:
            self.refresh_iterates(batch_size=100)

        frozen = GloroNet(
            epsilon=self.epsilon,
            model=self._f.model,
            num_iterations=0,
            hardcoded_kW=K.get_value(self._margin_layer._kW))

        frozen.trainable = False

        return frozen

    def refresh_iterates(
            self, 
            converge=True, 
            convergence_threshold=1e-4, 
            max_tries=1000, 
            batch_size=None,
            verbose=False):

        self._margin_layer.refresh_iterates(
            converge=converge,
            convergence_threshold=convergence_threshold,
            max_tries=max_tries,
            batch_size=batch_size,
            verbose=verbose)

        return self

    def update_iterates(self):
        self._margin_layer.update_iterates()

        return self

    def fit(self, *args, update_iterates=True, **kwargs):
        if update_iterates:
            includes_update_iterates = False

            if 'callbacks' in kwargs:
                callbacks = kwargs['callbacks']

                for callback in callbacks:
                    if isinstance(callback, UpdatePowerIterates):
                        includes_update_iterates = True

            else:
                callbacks = []

            if not includes_update_iterates:
                kwargs['callbacks'] = [
                    UpdatePowerIterates(verbose=False),
                    *callbacks
                ]

        return super().fit(*args, **kwargs)
        
    def get_config(self):
        config = super().get_config()
        config.update({
            'epsilon': self.epsilon,
            'model': self._f.model.get_config()
        })
        return config
      
    @classmethod
    def from_config(cls, config):
        model = Model.from_config(config['model'])
        
        return GloroNet(model=model, epsilon=config['epsilon'])
        
    class __ModelContainer(object):
        def __init__(self, model):
            self.model = model



import tensorflow as tf
import tensorflow.keras.backend as K

from tensorflow.keras.losses import CategoricalCrossentropy
from tensorflow.keras.losses import Loss
from tensorflow.keras.losses import SparseCategoricalCrossentropy

from gloro.training.utils import add_extra_column


class Crossentropy(Loss):

    def __init__(self, sparse=False, **kwargs):
        super().__init__()

        kwargs['from_logits'] = True

        if sparse:
            self._loss = SparseCategoricalCrossentropy(**kwargs)
        else:
            self._loss = CategoricalCrossentropy(**kwargs)

        self._sparse = sparse

    def call(self, y_true, y_pred):
        # Add an extra column of zeros for the bottom class.
        if not self._sparse:
            y_true = add_extra_column(y_true)

        return self._loss.call(y_true, y_pred)

    def get_config(self):
        config = self._loss.get_config()
        config.update({
            'sparse': self._sparse
        })
        return config


class Trades(Loss):

    def __init__(self, lam=1., sparse=False, robust_loss_string=None):
        super().__init__()

        self._lam = tf.Variable(lam, trainable=False)
        K.set_value(self._lam, lam)

        self._robust_loss_string = robust_loss_string

        if sparse:
            self._crossentropy = SparseCategoricalCrossentropy(from_logits=True)
        else:
            self._crossentropy = CategoricalCrossentropy(from_logits=True)

        if robust_loss_string is None:
            self._robust_loss = CategoricalCrossentropy(from_logits=False)

        else:
            self._robust_loss = tf.losses.get(robust_loss_string)

        self._sparse = sparse

    @property
    def lam(self):
        return K.get_value(self._lam)

    @lam.setter
    def lam(self, new_lam):
        K.set_value(self._lam, new_lam)

    def call(self, y_true, y_pred):
        # Add an extra column of zeros for the bottom class.
        if not self._sparse:
            y_true = add_extra_column(y_true)

        # Encourage predicting the correct class, even non-robustly.
        if self._sparse:
            standard_loss = self._crossentropy(y_true, y_pred[:, :-1])
        else:
            standard_loss = self._crossentropy(y_true[:, :-1], y_pred[:, :-1])

        # Encourage predicting robustly, even incorrectly. We take the robust
        # loss but using the model's prediction as the ground truth.
        new_ground_truth = add_extra_column(tf.nn.softmax(y_pred[:, :-1]))

        robust_loss = self._robust_loss(new_ground_truth, tf.nn.softmax(y_pred))

        return standard_loss + self._lam * robust_loss

    def get_config(self):
        return {
            'lam': self.lam,
            'robust_loss_string': self._robust_loss_string,
            'sparse': self._sparse
        }



from gloro import GloroNet
from gloro.training.losses import Crossentropy
from gloro.training.losses import Trades
from gloro.training.metrics import clean_acc
from gloro.training.metrics import vra


GLORO_CUSTOM_OBJECTS = {
    'GloroNet': GloroNet,
    'Crossentropy': Crossentropy,
    'Trades': Trades,
    'clean_acc': clean_acc,
    'vra': vra,
}



import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K

from tensorflow.keras.layers import Layer

from gloro.utils import global_lipschitz_bound_spectral_power
from gloro.utils import np_truncated_normal


class LipschitzMargin(Layer):
    def __init__(
            self, 
            epsilon, 
            model=None,
            num_iterations=5, 
            norm='l2', 
            maintain_state=True,
            hardcoded_kW=None, 
            **kwargs):

        super().__init__(**kwargs)

        self._epsilon = tf.Variable(epsilon, trainable=False)
        self._num_iterations = tf.Variable(num_iterations, trainable=False)

        self._norm = norm
        self._maintain_state = maintain_state if hardcoded_kW is None else False

        if model is None and hardcoded_kW is None:
            raise ValueError('must provide either `model` or `hardcoded_kW`')

        self._model = model
        self._hardcoded_kW = hardcoded_kW

    @property
    def epsilon(self):
        return K.get_value(self._epsilon)

    @epsilon.setter
    def epsilon(self, new_epsilon):
        K.set_value(self._epsilon, new_epsilon)

    @property
    def num_iterations(self):
        return K.get_value(self._num_iterations)

    @num_iterations.setter
    def num_iterations(self, new_num_iterations):
        K.set_value(self._num_iterations, new_num_iterations)

    @property
    def norm(self):
        return self._norm

    @property
    def maintain_state(self):
        return self._maintain_state

    @property
    def non_trainable_weights(self):
        return [self._epsilon, self._num_iterations]

    def call(self, y):
        if self._hardcoded_kW is None:
            k = global_lipschitz_bound_spectral_power(
                self._model, 
                self._num_iterations, 
                self._norm, 
                maintain_state=self._maintain_state)

            if self._maintain_state:
                k, iterates, updates = k

                self._iterates = iterates

                if not tf.executing_eagerly():
                    self._update_fn = K.function([], [k], updates=updates)

                else:
                    self._update_fn = K.function([], [k])

            # This is the L2 norm of the Lipschitz constant of the penultimate 
            # layer of the network multiplied by the weights of the penulimate 
            # layer.
            kW = k * self._model.layers[-1].kernel

        else:
            kW = self._hardcoded_kW

        self._kW = kW

        j = tf.argmax(y, axis=1)
        y_j = tf.reduce_max(y, axis=1, keepdims=True)
        where_not_j = tf.not_equal(y, y_j)

        # Get the weight column of the predicted class.
        kW_j = tf.gather(tf.transpose(kW), j)

        # Get weights that predict the value y_j - y_i for all i != j.
        kW_ij = kW_j[:,:,None] - kW[None]
        
        # We do this instead of `tf.linalg.norm(W_d)` because of an apparent bug
        # in `tf.linalg.norm` that leads to NaN values.
        K_ij = tf.sqrt(tf.reduce_sum(kW_ij * kW_ij, axis=1) + 1.e-10)

        # Save the margin Lipschitz constant for future reference. Put -1 in the 
        # position of j to signify that this Lipschitz constant is not defined
        # in this position.
        self._lipschitz_constant_tensor = tf.where(
            tf.equal(y, y_j), 
            tf.zeros_like(K_ij) - 1., 
            K_ij)

        y_bot_i = y + self._epsilon * K_ij

        # `y_bot_i` will be zero at the position of class j. However, we don't 
        # want to consider this class, so we replace the zero with negative
        # infinity so that when we find the maximum component for `y_bot_i` we 
        # don't get zero as a result of all of the components we care aobut 
        # being negative.
        y_bot_i = tf.where(
            tf.equal(y, y_j), 
            -np.infty + tf.zeros_like(y_bot_i), 
            y_bot_i)

        y_bot = tf.reduce_max(y_bot_i, axis=1, keepdims=True)

        return tf.concat([y, y_bot], axis=1)

    def refresh_iterates(
            self, 
            converge=True, 
            convergence_threshold=1e-4, 
            max_tries=1000, 
            batch_size=None,
            verbose=False):

        if not self._maintain_state:
            return self

        K.batch_set_value([
            (iterate, np_truncated_normal(K.int_shape(iterate)))
            for iterate in self._iterates
        ])

        if converge:
            if batch_size:
                old_num_iterations = self.num_iterations
                self.num_iterations = batch_size

            i = 0
            k_prev = None
            k = self._update_fn([])[0]

            while i < max_tries and (
                    k_prev is None or abs(k - k_prev) > convergence_threshold):

                k_prev = k
                k = self._update_fn([])[0]
                i += 1

            if batch_size:
                self.num_iterations = old_num_iterations

            if verbose:
                print(f'power method converged in {i+1} iterations')

        return self

    def update_iterates(self):
        if not self._maintain_state:
            return self

        self._update_fn([])

        return self

    def get_config(self):
        return {
            'epsilon': self.epsilon, 
            'num_iterations': self.num_iterations,
            'norm': self.norm,
            'name': self.name,

            # When saving, we freeze the Lipschitz constant so we don't have to
            # keep track of the model.
            'model': None,
            'hardcoded_kW': K.get_value(self._kW),
            'maintain_state': False,
        }



import numpy as np

from tensorflow.keras.callbacks import Callback
from tensorflow.keras.callbacks import LearningRateScheduler

import gloro

from gloro.training.losses import Trades


def _print_if_verbose(verbose):
    if verbose:
        return lambda s: print(s)

    else:
        return lambda s: None


def _check_is_gloro_net(callback, model):
    if not isinstance(model, gloro.GloroNet):
        raise ValueError(
            f'`{callback.__class__.__name__}` can only be used with a '
            f'`GloroNet` model')


def _parse_schedule_string(schedule_string, duration):
    if schedule_string == 'fixed':
        return [1.]

    elif schedule_string == 'linear_half':
        return [
            i / (duration // 2) for i in range(duration // 2 + 1)
        ]

    elif schedule_string == 'linear':
        return [
            i / (duration - 1) for i in range(duration)
        ]

    elif schedule_string == 'logarithmic':
        initial = 0.01
        final = 1.

        return [
            np.log(
                (np.exp(final) - np.exp(initial)) / (duration - 1) * i +
                np.exp(initial))
            for i in range(duration)
        ]

    elif schedule_string == 'logarithmic_half':
        initial = 0.01
        final = 1.

        return [
             np.log(
                (np.exp(final) - np.exp(initial)) / (duration // 2) * i +
                np.exp(initial))
            for i in range(duration // 2 + 1)
        ]

    else:
        raise ValueError(f'unrecognized schedule string: {schedule_string}')


class EpsilonScheduler(Callback):

    def __init__(self, schedule_string='fixed', verbose=True):
        super().__init__()

        self._schedule_string = schedule_string
        self._print = _print_if_verbose(verbose)

    def on_train_begin(self, logs=None):
        _check_is_gloro_net(self, self.model)

        schedule = np.array(_parse_schedule_string(
            self._schedule_string, self.params['epochs']))

        # We assume the epsilon originally set on the model is the final epsilon
        # we will train with, and scale the schedule accordingly.
        self._epoch_to_eps = self.model.epsilon * schedule

        self.__prev_eps = None
        self.__current_eps = self._epoch_to_eps[0]

    def on_epoch_begin(self, epoch, logs=None):
        if epoch < len(self._epoch_to_eps):
            next_eps = self._epoch_to_eps[epoch]

            self.model.epsilon = next_eps
        
            if self.__prev_eps is None or self.__prev_eps != self.__current_eps:
                self._print(f'---- setting epsilon={next_eps:.3f} ----')

            self.__prev_eps = self.__current_eps
            self.__current_eps = next_eps


class TradesScheduler(Callback):

    def __init__(self, schedule_string='fixed', verbose=True):
        super().__init__()

        self._schedule_string = schedule_string
        self._print = _print_if_verbose(verbose)

    def on_train_begin(self, logs=None):
        # Make sure that the model is actually using TRADES loss.
        if not isinstance(self.model.loss, Trades):
            raise ValueError(
                '`TradesScheduler` can only be used with `Trades` loss.')

        schedule = np.array(_parse_schedule_string(
            self._schedule_string, self.params['epochs']))

        # We assume the TRADES parameter originally set on the loss is the
        # desired final TRADES parameter, and scale the schedule accordingly.
        self._epoch_to_lam = self.model.loss.lam * schedule

        self.__prev_lam = None
        self.__current_lam = self._epoch_to_lam[0]

    def on_epoch_begin(self, epoch, logs=None):
        if epoch < len(self._epoch_to_lam):
            next_lam = self._epoch_to_lam[epoch]

            self.model.loss.lam = next_lam
        
            if self.__prev_lam is None or self.__prev_lam != self.__current_lam:
                self._print(
                    f'---- setting TRADES parameter={next_lam:.3f} ----')

            self.__prev_lam = self.__current_lam
            self.__current_lam = next_lam


class UpdatePowerIterates(Callback):
    def __init__(
            self, 
            convergence_threshold=1e-4, 
            iteration_batch_size=100, 
            verbose=True):

        super().__init__()

        self._convergence_threshold = convergence_threshold
        self._batch_size = iteration_batch_size

        self._print = _print_if_verbose(verbose)
        self._verbose = verbose

    def on_train_begin(self, logs=None):
        _check_is_gloro_net(self, self.model)

    def on_epoch_begin(self, epoch, logs=None):
        self._print('---- refreshing iterates ----')

        self.model.refresh_iterates(
            convergence_threshold=self._convergence_threshold,
            batch_size=self._batch_size,
            verbose=self._verbose)

    def on_train_end(self, logs=None):
        self._print('---- refreshing iterates ----')

        self.model.refresh_iterates(
            convergence_threshold=self._convergence_threshold,
            batch_size=self._batch_size,
            verbose=self._verbose)

    def on_train_batch_end(self, batch, logs=None):
        self.model.update_iterates()


class LrScheduler(LearningRateScheduler):

    def __init__(self, schedule_string, duration, verbose=True):
        if schedule_string == 'fixed':
            def scheduler(epoch, lr):
                return lr

        elif schedule_string.startswith('decay_to_'):
            end_lr = float(schedule_string.split('decay_to_')[1].split('_')[0])

            if schedule_string.endswith('_after_half'):
                def scheduler(epoch, lr):
                    if epoch < duration // 2:
                        self._initial_lr = lr

                        return lr

                    else:
                        return self._initial_lr * (self._initial_lr / end_lr)**(
                            -(epoch - duration // 2) / (duration // 2))

            else:
                def scheduler(epoch, lr):
                    if epoch == 0:
                        self._initial_lr = lr

                    return self._initial_lr * (self._initial_lr / end_lr)**(
                        -epoch / duration)

        elif (schedule_string.startswith('half_') and 
                schedule_string.endswith('_times')):

            times = int(schedule_string.split('half_')[1].split('_times')[0])
            period = duration // times

            def scheduler(epoch, lr):
                if epoch % period == period - 1:
                    return lr / 2.
                else:
                    return lr

        else:
            raise ValueError(f'unrecognized schedule string: {schedule_string}')

        super().__init__(scheduler, verbose=1 if verbose else 0)



import numpy as np
import tensorflow as tf


def add_extra_column(y):
    if isinstance(y, np.ndarray):
        return add_extra_column_np(y)
    else:
        return add_extra_column_tf(y)

def add_extra_column_tf(y):
    return tf.concat((y, tf.zeros((tf.shape(y)[0], 1))), axis=1)

def add_extra_column_np(y):
    return np.concatenate((y, np.zeros((y.shape[0], 1))), axis=1)



__version__ = '0.0.5'

# Disable eager execution if using Tensorflow 2.
try:
    from tensorflow.python.framework.ops import disable_eager_execution

    disable_eager_execution()
except:
    pass

from gloro.models import GloroNet

from gloro.constants import GLORO_CUSTOM_OBJECTS